precision highp float;

struct DirLight {
    vec3 direction;

    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

varying vec2 u_TexCoord;
varying vec3 u_FragPos;
varying vec3 u_Normal;

uniform int lightMode;
uniform vec3 viewPos;
uniform DirLight dirLight;
uniform float shininess;

uniform sampler2D texture;
uniform sampler2D baseTexture;
uniform sampler2D maskTexture;
uniform float alpha;

uniform int blendMode;
uniform int maskMode;
uniform vec2 viewportSize;

vec3 calcDirLight(vec4 src, DirLight light, vec3 normal, vec3 viewDir);
vec4 calcLight(vec4 src);

vec4 mask(vec4 src, vec2 maskUV) {
    vec4 color = src;
    if (maskMode != -1) {
        vec4 maskColor = texture2D(maskTexture, maskUV);
        if (maskMode < 2) {
            if (maskMode == 0) {
                color *= maskColor.a;
            } else {
                color *= (1.0 - maskColor.a);
            }
        } else {
            if (maskMode == 2) {
                color *= dot(maskColor.rgb, vec3(0.2125, 0.7154, 0.0721));
            } else {
                color *= (1.0 - dot(maskColor.rgb, vec3(0.2125, 0.7154, 0.0721)));
            }
        }
    }
    return color;
}


/* blend functions */

vec4 blendNormal(vec4 base, vec4 overlay) { // 0
    return base + overlay - base * overlay.a;
}

vec4 blendScreen(vec4 base, vec4 overlay) { // 1
    vec4 whiteColor = vec4(1.0);
    return whiteColor - ((whiteColor - overlay) * (whiteColor - base));
}

vec4 blendMultiply(vec4 base, vec4 overlay) { // 2
    return overlay * base + overlay * (1.0 - base.a) + base * (1.0 - overlay.a);
}

vec4 blendOverlay(vec4 base, vec4 overlay) { // 3
    float ra;
    if (2.0 * base.r < base.a) {
        ra = 2.0 * overlay.r * base.r + overlay.r * (1.0 - base.a) + base.r * (1.0 - overlay.a);
    } else {
        ra = overlay.a * base.a - 2.0 * (base.a - base.r) * (overlay.a - overlay.r) + overlay.r * (1.0 - base.a) + base.r * (1.0 - overlay.a);
    }

    float ga;
    if (2.0 * base.g < base.a) {
        ga = 2.0 * overlay.g * base.g + overlay.g * (1.0 - base.a) + base.g * (1.0 - overlay.a);
    } else {
        ga = overlay.a * base.a - 2.0 * (base.a - base.g) * (overlay.a - overlay.g) + overlay.g * (1.0 - base.a) + base.g * (1.0 - overlay.a);
    }

    float ba;
    if (2.0 * base.b < base.a) {
        ba = 2.0 * overlay.b * base.b + overlay.b * (1.0 - base.a) + base.b * (1.0 - overlay.a);
    } else {
        ba = overlay.a * base.a - 2.0 * (base.a - base.b) * (overlay.a - overlay.b) + overlay.b * (1.0 - base.a) + base.b * (1.0 - overlay.a);
    }

    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4(ra, ga, ba, alpha);
}

vec4 blendDarken(vec4 base, vec4 overlay) { // 4
    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4(min(overlay.rgb * base.a, base.rgb * overlay.a) + overlay.rgb * (1.0 - base.a) + base.rgb * (1.0 - overlay.a), alpha);
}

vec4 blendLighten(vec4 base, vec4 overlay) { // 5
    return max(base, overlay);
}

vec4 blendLightDodge(vec4 base, vec4 overlay) { // 6
    vec3 baseOverlayAlphaProduct = vec3(overlay.a * base.a);
    vec3 rightHandProduct = overlay.rgb * (1.0 - base.a) + base.rgb * (1.0 - overlay.a);

    vec3 firstBlendColor = baseOverlayAlphaProduct + rightHandProduct;
    vec3 overlayRGB = clamp((overlay.rgb / clamp(overlay.a, 0.01, 1.0)) * step(0.0, overlay.a), 0.0, 0.99);

    vec3 secondBlendColor = (base.rgb * overlay.a) / (1.0 - overlayRGB) + rightHandProduct;
    vec3 colorChoice = step((overlay.rgb * base.a + base.rgb * overlay.a), baseOverlayAlphaProduct);

    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4(mix(firstBlendColor, secondBlendColor, colorChoice), alpha);
}

vec4 blendColorBurn(vec4 base, vec4 overlay) { // 7
    float alpha = base.a + overlay.a - base.a * overlay.a;
    float CbCsAlphaProduct = base.a * overlay.a;

    float ra;
    if (overlay.r == 0.0) {
        ra = 0.0;
    } else {
        ra = CbCsAlphaProduct - min(CbCsAlphaProduct, (CbCsAlphaProduct * overlay.a - base.r * overlay.a * overlay.a) / overlay.r);
    }

    float rg;
    if (overlay.g == 0.0) {
        rg = 0.0;
    } else {
        rg = CbCsAlphaProduct - min(CbCsAlphaProduct, (CbCsAlphaProduct * overlay.a - base.g * overlay.a * overlay.a) / overlay.g);
    }

    float rb;
    if (overlay.b == 0.0) {
        rb = 0.0;
    } else {
        rb = CbCsAlphaProduct - min(CbCsAlphaProduct, (CbCsAlphaProduct * overlay.a - base.b * overlay.a * overlay.a) / overlay.b);
    }

    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + vec3(ra, rg, rb), alpha);
}

vec4 blendHardLight(vec4 base, vec4 overlay) { // 8
    highp float ra;
    if (2.0 * overlay.r < overlay.a) {
        ra = 2.0 * overlay.r * base.r + overlay.r * (1.0 - base.a) + base.r * (1.0 - overlay.a);
    } else {
        ra = overlay.a * base.a - 2.0 * (base.a - base.r) * (overlay.a - overlay.r) + overlay.r * (1.0 - base.a) + base.r * (1.0 - overlay.a);
    }

    highp float ga;
    if (2.0 * overlay.g < overlay.a) {
        ga = 2.0 * overlay.g * base.g + overlay.g * (1.0 - base.a) + base.g * (1.0 - overlay.a);
    } else {
        ga = overlay.a * base.a - 2.0 * (base.a - base.g) * (overlay.a - overlay.g) + overlay.g * (1.0 - base.a) + base.g * (1.0 - overlay.a);
    }

    highp float ba;
    if (2.0 * overlay.b < overlay.a) {
        ba = 2.0 * overlay.b * base.b + overlay.b * (1.0 - base.a) + base.b * (1.0 - overlay.a);
    } else {
        ba = overlay.a * base.a - 2.0 * (base.a - base.b) * (overlay.a - overlay.b) + overlay.b * (1.0 - base.a) + base.b * (1.0 - overlay.a);
    }

    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4(ra, ga, ba, alpha);
}

float eachcomponentSoftLight(float Cb, float Cs) {
    if (Cs <= 0.5) {
        return Cb - (1.0 - 2.0 * Cs) * Cb * (1.0 - Cb);
    } else {
        float zeus = 0.0;
        if (Cb <= 0.25) {
            zeus = ((16.0 * Cb - 12.0) * Cb + 4.0) * Cb;
        } else {
            zeus = sqrt(Cb);
        }
        return Cb + (2.0 * Cs - 1.0) * (zeus - Cb);
    }
}

vec4 blendSoftLight(vec4 base, vec4 overlay) { // 9
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    float r = eachcomponentSoftLight(baseRGB.r, overlayRGB.r);
    float g = eachcomponentSoftLight(baseRGB.g, overlayRGB.g);
    float b = eachcomponentSoftLight(baseRGB.b, overlayRGB.b);
    vec3 color = clamp(vec3(r, g, b), 0.0, 1.0);
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendDifference(vec4 base, vec4 overlay) { // 10
    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4(abs(overlay.rgb - base.rgb), alpha);
}

vec4 blendExclusion(vec4 base, vec4 overlay) { // 11
    float alpha = overlay.a + base.a - overlay.a * base.a;
    return vec4((overlay.rgb * base.a + base.rgb * overlay.a - 2.0 * overlay.rgb * base.rgb) + overlay.rgb * (1.0 - base.a) + base.rgb * (1.0 - overlay.a), alpha);
}

vec4 blendHardMix(vec4 base, vec4 overlay) { // 12
    float plato = base.a * overlay.a;
    vec3 color = step(plato, overlay.rgb * base.a + base.rgb * overlay.a) * (1.0 - step(plato, 0.0));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendDivide(vec4 base, vec4 overlay) { // 13
    float odin = 1.0;
    if (base.a > 0.0) {
        odin = overlay.a / base.a;
    }

    float ra = 1.0 - step(base.r, 0.0);
    if (overlay.r > 0.0) {
        ra = base.r / overlay.r * odin;
    }

    float rg = 1.0 - step(base.g, 0.0);
    if (overlay.g > 0.0) {
        rg = base.g / overlay.g * odin;
    }

    float rb = 1.0 - step(base.b, 0.0);
    if (overlay.b > 0.0) {
        rb = base.b / overlay.b * odin;
    }

    vec3 color = clamp(vec3(ra, rg, rb), 0.0, 1.0);
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

float eachcomponentVividLight(float Cb, float Cs, float Ab, float As) {
    float nonPremultiplyCb = Cb / (Ab + step(Ab, 0.0));
    float nonPremultiplyCs = Cs / (As + step(As, 0.0));
    if (nonPremultiplyCs <= 0.5) {
        if (nonPremultiplyCs == 0.0) {
            return 0.0;
        } else {
            return 1.0 - (1.0 - nonPremultiplyCb) / (2.0 * nonPremultiplyCs);
        }
    } else {
        if (nonPremultiplyCs == 1.0) {
            return 1.0;
        } else {
            return nonPremultiplyCb / (2.0 * (1.0 - nonPremultiplyCs));
        }
    }
}

vec4 blendVividLight(vec4 base, vec4 overlay) { // 14
    float r = eachcomponentVividLight(base.r, overlay.r, base.a, overlay.a);
    float g = eachcomponentVividLight(base.g, overlay.g, base.a, overlay.a);
    float b = eachcomponentVividLight(base.b, overlay.b, base.a, overlay.a);
    vec3 color = clamp(vec3(r, g, b), 0.0, 1.0);
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendDarker(vec4 base, vec4 overlay) { // 15
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    float baseSum = baseRGB.r * 0.299 + baseRGB.g * 0.587 + baseRGB.b * 0.114;
    float overlaySum = overlayRGB.r * 0.299 + overlayRGB.g * 0.587 + overlayRGB.b * 0.114;
    vec3 color = overlaySum > baseSum ? overlayRGB : baseRGB;
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendLighter(vec4 base, vec4 overlay) { // 16
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    float baseSum = baseRGB.r * 0.299 + baseRGB.g * 0.587 + baseRGB.b * 0.114;
    float overlaySum = overlayRGB.r * 0.299 + overlayRGB.g * 0.587 + overlayRGB.b * 0.114;
    vec3 color = overlaySum > baseSum ? overlayRGB : baseRGB;
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendLinearBurn(vec4 base, vec4 overlay) { // 17
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * clamp(baseRGB + overlayRGB - 1.0, 0.0, 1.0), base.a + overlay.a - base.a * overlay.a);
}

vec4 blendLinearDodge(vec4 base, vec4 overlay) { // 18
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * clamp(baseRGB + overlayRGB, 0.0, 1.0), base.a + overlay.a - base.a * overlay.a);
}

vec4 blendSubtract(vec4 base, vec4 overlay) { // 19
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    vec3 color = max(baseRGB - overlayRGB, 0.0);
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendLinearLight(vec4 base, vec4 overlay) { // 20
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * clamp(baseRGB + 2.0 * overlayRGB - 1.0, 0.0, 1.0), base.a + overlay.a - base.a * overlay.a);
}

float eachcomponentPinLight(float Cb, float Cs) {
    if (Cs <= 0.5) {
        return min(2.0 * Cs, Cb);
    } else {
        return max(2.0 * (Cs - 0.5), Cb);
    }
}

vec4 blendPinLight(vec4 base, vec4 overlay) { // 21
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    float r = eachcomponentPinLight(baseRGB.r, overlayRGB.r);
    float g = eachcomponentPinLight(baseRGB.g, overlayRGB.g);
    float b = eachcomponentPinLight(baseRGB.b, overlayRGB.b);
    vec3 color = vec3(r, g, b);
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

float lum(vec3 c) {
    return dot(c, vec3(0.3, 0.59, 0.11));
}

vec3 clipcolor(vec3 c) {
    float l = lum(c);
    float n = min(min(c.r, c.g), c.b);
    float x = max(max(c.r, c.g), c.b);
    if (n < 0.0) {
        c.r = l + ((c.r - l) * l) / (l - n);
        c.g = l + ((c.g - l) * l) / (l - n);
        c.b = l + ((c.b - l) * l) / (l - n);
    }
    if (x > 1.0) {
        c.r = l + ((c.r - l) * (1.0 - l)) / (x - l);
        c.g = l + ((c.g - l) * (1.0 - l)) / (x - l);
        c.b = l + ((c.b - l) * (1.0 - l)) / (x - l);
    }
    return c;
}

vec3 setlum(vec3 c, float l) {
    float d = l - lum(c);
    c = c + vec3(d);
    return clipcolor(c);
}

float sat(vec3 c) {
    float n = min(min(c.r, c.g), c.b);
    float x = max(max(c.r, c.g), c.b);
    return x - n;
}

float mid(float cmin, float cmid, float cmax, float s) {
    return ((cmid - cmin) * s) / (cmax - cmin);
}

lowp vec3 setsat(vec3 c, float s) {
    if (c.r > c.g) {
        if (c.r > c.b) {
            if (c.g > c.b) {
                c.g = mid(c.b, c.g, c.r, s);
                c.b = 0.0;
            } else {
                c.b = mid(c.g, c.b, c.r, s);
                c.g = 0.0;
            }
            c.r = s;
        } else {
            c.r = mid(c.g, c.r, c.b, s);
            c.b = s;
            c.g = 0.0;
        }
    } else if (c.r > c.b) {
        c.r = mid(c.b, c.r, c.g, s);
        c.g = s;
        c.b = 0.0;
    } else if (c.g > c.b) {
        c.b = mid(c.r, c.b, c.g, s);
        c.g = s;
        c.r = 0.0;
    } else if (c.b > c.g) {
        c.g = mid(c.r, c.g, c.b, s);
        c.b = s;
        c.r = 0.0;
    } else {
        c = vec3(0.0);
    }
    return c;
}

vec4 blendHue(vec4 base, vec4 overlay) { // 22
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    vec3 color = setlum(setsat(overlayRGB, sat(baseRGB)), lum(baseRGB));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendSaturation(vec4 base, vec4 overlay) { // 23
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    vec3 color = setlum(setsat(baseRGB, sat(overlayRGB)), lum(baseRGB));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendColor(vec4 base, vec4 overlay) { // 24
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    vec3 color = setlum(overlayRGB, lum(baseRGB));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}

vec4 blendLuminosity(vec4 base, vec4 overlay) { // 25
    vec3 baseRGB = base.rgb / (base.a + step(base.a, 0.0));
    vec3 overlayRGB = overlay.rgb / (overlay.a + step(overlay.a, 0.0));
    vec3 color = setlum(baseRGB, lum(overlayRGB));
    return vec4((1.0 - overlay.a) * base.rgb + (1.0 - base.a) * overlay.rgb + base.a * overlay.a * color, base.a + overlay.a - base.a * overlay.a);
}


void main() {
    vec2 gUV = gl_FragCoord.xy / viewportSize;
    vec4 lightR = calcLight(texture2D(texture, u_TexCoord));
    vec4 result = mask(lightR, gUV) * alpha;
    if (result.a < 0.01) {
        discard;
    }
    if (blendMode < 13) {
        if (blendMode < 6) {
            if (blendMode < 3) {
                if (blendMode < 1) {
                    if (blendMode == 0) {
                        gl_FragColor = blendNormal(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = result;
                    }
                } else {
                    if (blendMode == 1) {
                        gl_FragColor = blendScreen(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = blendMultiply(texture2D(baseTexture, gUV), result);
                    }
                }
            } else {
                if (blendMode == 3) {
                    gl_FragColor = blendOverlay(texture2D(baseTexture, gUV), result);
                } else if (blendMode == 4) {
                    gl_FragColor = blendDarken(texture2D(baseTexture, gUV), result);
                } else {
                    gl_FragColor = blendLighten(texture2D(baseTexture, gUV), result);
                }
            }
        } else {
            if (blendMode < 10) {
                if (blendMode < 8) {
                    if (blendMode == 6) {
                        gl_FragColor = blendLightDodge(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = blendColorBurn(texture2D(baseTexture, gUV), result);
                    }
                } else {
                    if (blendMode == 8) {
                        gl_FragColor = blendHardLight(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = blendSoftLight(texture2D(baseTexture, gUV), result);
                    }
                }
            } else {
                if (blendMode == 10) {
                    gl_FragColor = blendDifference(texture2D(baseTexture, gUV), result);
                } else if (blendMode == 11) {
                    gl_FragColor = blendExclusion(texture2D(baseTexture, gUV), result);
                } else {
                    gl_FragColor = blendHardMix(texture2D(baseTexture, gUV), result);
                }
            }
        }
    } else {
        if (blendMode < 20) {
            if (blendMode < 17) {
                if (blendMode < 15) {
                    if (blendMode == 13) {
                        gl_FragColor = blendDivide(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = blendVividLight(texture2D(baseTexture, gUV), result);
                    }
                } else {
                    if (blendMode == 15) {
                        gl_FragColor = blendDarker(texture2D(baseTexture, gUV), result);
                    } else {
                        gl_FragColor = blendLighter(texture2D(baseTexture, gUV), result);
                    }
                }
            } else {
                if (blendMode == 17) {
                    gl_FragColor = blendLinearBurn(texture2D(baseTexture, gUV), result);
                } else if (blendMode == 18) {
                    gl_FragColor = blendLinearDodge(texture2D(baseTexture, gUV), result);
                } else {
                    gl_FragColor = blendSubtract(texture2D(baseTexture, gUV), result);
                }
            }
        } else {
            if (blendMode < 23) {
                if (blendMode == 20) {
                    gl_FragColor = blendLinearLight(texture2D(baseTexture, gUV), result);
                } else if (blendMode == 21) {
                    gl_FragColor = blendPinLight(texture2D(baseTexture, gUV), result);
                } else {
                    gl_FragColor = blendHue(texture2D(baseTexture, gUV), result);
                }
            } else {
                if (blendMode == 23) {
                    gl_FragColor = blendSaturation(texture2D(baseTexture, gUV), result);
                } else if (blendMode == 24) {
                    gl_FragColor = blendColor(texture2D(baseTexture, gUV), result);
                } else {
                    gl_FragColor = blendLuminosity(texture2D(baseTexture, gUV), result);
                }
            }
        }
    }

}

vec4 calcLight(vec4 src) {
    if (lightMode == 1) {
        vec3 norm = normalize(u_Normal);
        vec3 viewDir = normalize(viewPos - u_FragPos);
        return vec4(calcDirLight(src, dirLight, norm, viewDir), src.a);
    }
    return src;
}

// calculates the color when using a directional light.
vec3 calcDirLight(vec4 src, DirLight light, vec3 normal, vec3 viewDir) {
    vec3 lightDir = normalize(-light.direction);
    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0);
    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), shininess);
    // combine results
    vec3 tex = vec3(src);
    vec3 ambient = light.ambient * tex;
    vec3 diffuse = light.diffuse * diff * tex;
    vec3 specular = light.specular * spec * tex;
    return (ambient + diffuse + specular);
}


